/* MoleculesXYZDataset.java
 * Author: Nina Jeliazkova
 * Date: Sep 29, 2006 
 * Revision: 0.1 
 * 
 * Copyright (C) 2005-2006  Nina Jeliazkova
 * 
 * Contact: nina@acad.bg
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1
 * of the License, or (at your option) any later version.
 * All we ask is that proper credit is given for our work, which includes
 * - but is not limited to - adding the above copyright notice to the beginning
 * of your source code files, and to any copyright notice that you may distribute
 * with programs based on this work.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 */

package chemCast.ui;

import java.awt.Color;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.text.NumberFormat;

import org.jfree.chart.JFreeChart;
import org.jfree.data.general.SeriesChangeEvent;
import org.jfree.data.xy.AbstractXYZDataset;

import ambit.data.molecule.IAtomContainersList;
import chemCast.exception.SubsetException;
import chemCast.similarity.data.Subset;
import chemCast.similarity.data.Subsets;

/**
 * 
 * A {@link JFreeChart} dataset, allowing to  draw scatterplot of arbitrary properties of a  {@link Subset}
 * @author Nina Jeliazkova nina@acad.bg
 * <b>Modified</b> Oct 28, 2006
 */
public class MoleculesXYZDataset extends AbstractXYZDataset implements PropertyChangeListener {
    /**
     * 
     */
    private static final long serialVersionUID = -8555431723337857566L;
    protected Subsets subsets;
    protected Object propertyX;
    protected Object propertyY;
    protected Object propertyZ;
    protected Object labelProperty;
    protected Object tooltipProperty;
    protected Object colorProperty;
    protected static String blank = "";
    protected boolean enabled = true;
    protected NumberFormat nf = NumberFormat.getInstance();
    //protected ArrayList<Number[][]> data;
    
    public  MoleculesXYZDataset(Subsets subsets, 
            Object propertyX, Object propertyY, Object propertyZ, 
            Object labelProperty, Object tooltipProperty, Object colorProperty) {
        super();
        //data = new ArrayList<Number[][]>();
        setSubsets(subsets);
        setPropertyX(propertyX);
        setPropertyY(propertyY);
        setPropertyZ(propertyZ);
        
        setLabelProperty(labelProperty);
        setTooltipProperty(tooltipProperty);
        setColorProperty(colorProperty);
        nf.setMaximumFractionDigits(3);
    }
    /**
     * e.g. "CAS #"
     * @param series
     * @param item
     * @return
     */
    public String getLabel(int series, int item) {
        if (subsets == null) return blank;
    	if (labelProperty != null) {
            try {
                Object o = subsets.get(series).getData().getProperty(item,labelProperty);
                if (o != null) {
                    if (o instanceof Number) {
                        return nf.format(o);
                    } else return o.toString();
                }
            } catch (SubsetException x) {
                return blank;
            }
    	} 
    	return blank;
    }
    /**
     * e.g. "CHEMICAL NAME" 
     * @param series
     * @param item
     * @return
     */
    public String getTooltip(int series,int item) {
        if (subsets == null) return blank;
        if (subsets.get(series)==null) return blank;
    	if (tooltipProperty != null) 
            try {
                
                IAtomContainersList l = subsets.get(series).getData();
                if (l ==null) return blank;
        		Object o = l.getProperty(item,tooltipProperty);
            	if ((o != null) && (!o.equals(blank))) return o.toString();
            } catch (SubsetException x) {
                return blank;
            }     
    	return getSeriesKey(series) + blank + '#' + Integer.toString(item+1);
        	
    }    
    protected Number getValue(Object property, int series, int item) {
        if (!enabled) return 0;
        if ((property == null) || ("".equals(property.toString()))) return 0;
        
        if (subsets == null) return 0;
        try {
            
            Object o = subsets.get(series).getData().getProperty(item,property);
            if (o==null) return 0;
            //System.out.println(o);
            if (o instanceof Number) 	return (Number)o;
            else 
                return new Double(o.toString());
        } catch (Exception x) {
            return 0;
        }
    }
    
    public Number getX(int series, int item) {
        /*
        Number[][] d = data.get(series);
        if (d != null) return d[item][0];
        else return 0;
        */
        return getValue(propertyX,series, item);
    }
    public Number getY(int series, int item) {
        /*
        Number[][] d = data.get(series);
        if (d != null) return d[item][1];
        else return 0;
        */
        return getValue(propertyY,series, item);
    }
    public Number getZ(int series, int item) {
        return getValue(propertyZ,series, item);
    }
    @Override
    public Comparable getSeriesKey(int series) {
        if ((subsets == null) || 
                (subsets.get(series) == null)) return blank;
        try {
            return subsets.get(series).toString();
        } catch (Exception x) {
        	return blank;
        }
    }
    public int getItemCount(int series) {
        if (!enabled) return 0;
        if (subsets == null) return 0;
        try {
            if (series>subsets.size()) return 0;
            if (subsets.get(series) == null) return 0;
            if (subsets.get(series).getData() == null) return 0;
            //System.out.println(subsets + " " + subsets.get(series).getData().getAtomContainerCount());
            return subsets.get(series).getData().getAtomContainerCount();
        } catch (Exception x) {
            x.printStackTrace();
            return 0;
        }
    }
    @Override
    public int getSeriesCount() {
        if (subsets == null) return 0;
        if (enabled)
            return subsets.size() ;
        else return 0;
    }
    
    public synchronized Object getPropertyX() {
        return propertyX;
    }
    public synchronized void setProperty(int index, Object property) {
        switch (index) {
        case 0: { setPropertyX(property); return;}
        case 1: { setPropertyY(property); return;}
        case 2: { setPropertyZ(property); return;}
        }
    }
    
    
    public synchronized void setPropertyX(Object propertyX) {
        this.propertyX = propertyX;
        bufferProperties();
        fireDatasetChanged();
    }
    public synchronized Object getPropertyY() {
        return propertyY;
    }
    public synchronized void setPropertyY(Object propertyY) {
        this.propertyY = propertyY;
        bufferProperties();
        fireDatasetChanged();
    }
    public synchronized Subsets getReaders() {
        return subsets;
    }
    public synchronized void setSubsets(Subsets subsets) {
        
        if (subsets == null) return;
        //System.out.println("setSubsets " + subsets);
        //data.clear();
    	if (this.subsets != null) {
    		this.subsets.removePropertyChangeListener(this);
    		for (int i=0; i < this.subsets.size(); i++)
    			this.subsets.get(i).removePropertyChangeListener(this);
    	}	
        this.subsets = subsets;
        if (subsets != null) {
        	subsets.addPropertyChangeListener(this);
			for (int i=0; i < subsets.size(); i++) {
				subsets.get(i).addPropertyChangeListener(this);
                /*
                try {
                    if (subsets.get(i).getData() != null) {
                        Number[][] values = new Number[subsets.get(i).getData().getAtomContainerCount()][2]; 
                        data.add(values);
                    } else data.add(null);
                } catch (Exception x) {
                    x.printStackTrace();
                    data.add(null);
                }
                */
            }    
        }
        
        fireDatasetChanged();
    }
    public synchronized Object getPropertyZ() {
        return propertyZ;
    }
    public synchronized void setPropertyZ(Object propertyZ) {
         
            this.propertyZ = propertyZ;
        fireDatasetChanged();
    }
	public Object getLabelProperty() {
		return labelProperty;
	}
	public void setLabelProperty(Object labelProperty) {
		this.labelProperty = labelProperty;
		   fireDatasetChanged();
	}
	public Object getTooltipProperty() {
		return tooltipProperty;
	}
	public void setTooltipProperty(Object tooltipProperty) {
		this.tooltipProperty = tooltipProperty;
		   fireDatasetChanged();
	}
	@Override
	public void seriesChanged(SeriesChangeEvent arg0) {
		super.seriesChanged(arg0);
		fireDatasetChanged();
	}
	public Color getColor(int series) {
		if (series < subsets.size())
            if (subsets.get(series) != null)
                return subsets.get(series).getColor();
		return Color.GRAY;
    }
    
    public boolean isVisible(int series) {
        try {
        return subsets.get(series).isEnabled();
        } catch (Exception x) {
            return true;
        }
    }
    public void setVisible() {
        fireDatasetChanged();

    }    
    public void propertyChange(PropertyChangeEvent evt) {
        if (evt.getPropertyName().equals(Subsets.p_groups))
            setSubsets((Subsets) evt.getNewValue());
        /*
        if (evt instanceof SelectItemEvent) {
            if (evt.getNewValue() == null) setSubsets(null);    
            else {
                setSubsets(((SelectedItem) evt.getNewValue()).getDataset());
            }
        }
        */
    	//System.out.println(getClass().getName() + " Propertyname " + evt.getPropertyName() + " New " + evt.getNewValue());
    	fireDatasetChanged();
    }
    public synchronized boolean isEnabled() {
        return enabled;
    }
    public synchronized void setEnabled(boolean enabled) {
        this.enabled = enabled;
        //this.enabled = false;
        fireDatasetChanged();
    }
    protected void bufferProperties() {
        return;
        /*
        if (subsets != null)
            for (int i=0; i < subsets.size();i++) {
                try {
                    IAtomContainersList list = subsets.get(i).getData();
                    if (i >= data.size()) 
                        if (list == null) {
                            data.add(null);
                            continue;
                        } else data.add(i, new Number[list.getAtomContainerCount()][2]);

                    Number[][] values = data.get(i);
                    if ((list!=null) && (values != null))
                        for (int j=0; j < values.length;j++) { 
                            Object o = list.getProperty(j, propertyX);
                            if ((o!=null) && (o instanceof Number)) values[j][0] = (Number)o;
                            else try { 
                                values[j][0] = new Double(o.toString());
                            } catch (Exception x) {
                                values[j][0] = 0;
                            }
                            o =list.getProperty(j, propertyY);
                            if ((o!=null) && (o instanceof Number)) values[j][1] = (Number)o;
                            else try { 
                                values[j][1] = new Double(o.toString());
                            } catch (Exception x) {
                                values[j][1] = 0;
                            }
                        }
                } catch (Exception x) {
                    x.printStackTrace();
                }
            }
            */  
    }
    @Override
    public String toString() {
        return subsets.toString();
    }
    public synchronized Object getColorProperty() {
        return colorProperty;
    }
    public synchronized void setColorProperty(Object colorProperty) {
        if ("NONE".equals(colorProperty)) this.colorProperty = "";
        else this.colorProperty = colorProperty;
        fireDatasetChanged();
    }
}
