/*
Copyright Ideaconsult Ltd. (C) 2005-2007 

Contact: nina@acad.bg

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.
All we ask is that proper credit is given for our work, which includes
- but is not limited to - adding the above copyright notice to the beginning
of your source code files, and to any copyright notice that you may distribute
with programs based on this work.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/
package toxTree.core;

import java.beans.XMLDecoder;
import java.io.File;
import java.io.FileFilter;
import java.io.IOException;
import java.io.InputStream;
import java.lang.reflect.Modifier;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;

import javax.swing.JOptionPane;

import toxTree.logging.TTLogger;

/**
 * 
 * A singleton class with static method to provide some sort of introspection :)
 * Used in decision tree editor to provide a list of available decision trees,
 * available rules, available categories. <br>
 * This is done by searching for classes implementing
 * {@link toxTree.core.IDecisionMethod}, {@link toxTree.core.IDecisionRule},
 * and {@link toxTree.core.IDecisionCategory} in all .jar files in a user
 * defined directory.
 * 
 * @author Nina Jeliazkova nina@acad.bg <b>Modified</b> 2005-10-18
 */
public class Introspection {
	protected static String[] defaultLocation = { ".", "dist","ext" };

	protected static TTLogger logger = new TTLogger(Introspection.class);

	protected static ToxTreeClassLoader loader;

	protected Introspection() {
		super();
		loader = null;
	}
    public static void addDirectory(String dirName) {
    	defaultLocation[0] = new String(dirName);
    	return;
        }
    public static String[] getDefaultDirectories() {
    	return defaultLocation;
    }	

	/*
	 * public static ArrayList getClassImplementing(String anInterface) {
	 * ArrayList result = new ArrayList(); Package[] packages =
	 * Package.getPackages();
	 * 
	 * for (int i =0; i<packages.length;i++) { Package self = packages[i]; if
	 * ((self != null) && (self.getName().startsWith("toxTree"))) {
	 *  } }
	 * 
	 * return result; }
	 */
	/**
	 * this is just a test for the concept of introspection
	 */
	public static void listBaseTypes(Class cls, String prefix) {
		if (cls == Object.class)
			return;
		Class[] itfs = cls.getInterfaces();

		for (int n = 0; n < itfs.length; n++) {
			System.out.println(prefix + "implements " + itfs[n]);
			listBaseTypes(itfs[n], prefix + "\t");
		}
		Class base = cls.getSuperclass();
		if (base == null)
			return;
		System.out.println(prefix + "extends " + base);
		listBaseTypes(base, prefix + "\t");
	}

	/**
	 * Verifies if a class implements an interface
	 * 
	 * @param className -
	 *            the name of the class to be verified
	 * @param interfaceName -
	 *            the name of the interface to be searched for
	 * @return Class
	 */
	public static Class implementsInterface(String className,
			String interfaceName) throws toxTree.exceptions.IntrospectionException {
		try {
			Class clazz = Class.forName(className);
			int modifier = clazz.getModifiers();
			if (Modifier.isAbstract(modifier))
				return null;
			else if (Modifier.isInterface(modifier))
				return null;
			return implementsInterface(clazz, interfaceName);
		} catch (ClassNotFoundException x) {
			//
			if (loader != null) {
				try {
					Class clazz = loader.loadClass(className);
					int modifier = clazz.getModifiers();
					if (Modifier.isAbstract(modifier))
						return null;
					else if (Modifier.isInterface(modifier))
						return null;
					return implementsInterface(clazz, interfaceName);
				} catch (ClassNotFoundException xx) {

					throw new toxTree.exceptions.IntrospectionException(xx);
				}
			}
			// x.printStackTrace();
			return null;
		} catch (Exception x) {
			throw new toxTree.exceptions.IntrospectionException(x);
		}
	}

	/**
	 * Verifies if a class implements an interface
	 * 
	 * @param clazz -
	 *            the class to be verified
	 * @param interfaceName -
	 *            the name of the interface to be searched for
	 * @return Class
	 */
	public static Class implementsInterface(Class clazz, String interfaceName) {
		Class[] interfaces = clazz.getInterfaces();

		for (int i = 0; i < interfaces.length; i++)
			if (interfaces[i].getName().equals(interfaceName))
				return clazz;

		// try base class
		Class base = clazz.getSuperclass();
		if (base == null)
			return null;
		else
			return implementsInterface(base, interfaceName);

	}

	/**
	 * Finds classes implementing an interface in all .jar files in a user
	 * defined directory.
	 * 
	 * @param directory -
	 *            the directory with jar files to be searched for
	 * @param interfaceName -
	 *            the name of the interface to be searched for
	 * @return ToxTreePackageEntries of String , each item is a class name
	 */

	public static ToxTreePackageEntries implementInterface(ClassLoader classLoader,
			File directory, String interfaceName) throws toxTree.exceptions.IntrospectionException {
		setLoader(loader);
		logger.info("Looking for .jar files in\t", directory
						.getAbsolutePath());
		File[] files = directory.listFiles(new FileFilter() {
			public boolean accept(File arg0) {
				//return arg0.getName().toLowerCase().endsWith(".jar");
			    return arg0.getName().toLowerCase().endsWith(".jar") && 
				arg0.getName().toLowerCase().startsWith("tox");				
			}
		});

		ToxTreePackageEntries rules = new ToxTreePackageEntries();
		if (files == null)
			return rules;
		for (int i = 0; i < files.length; i++)

			try {
				logger.info("Inspecting jar file\t", files[i], "\t", files[i]
						.toURL());
				JarFile jar = new JarFile(files[i]);
				Enumeration entries = jar.entries();
				if (loader != null)
					loader.addURL(files[i].toURL());
				while (entries.hasMoreElements()) {
					JarEntry entry = (JarEntry) entries.nextElement();
					if (!entry.getName().endsWith("class"))
						continue;

					String name = entry.getName();

					name = name.replaceAll("/", ".").substring(0,
							name.indexOf(".class"));

					Class rule = Introspection.implementsInterface(name,
							interfaceName);

					if (rule != null) {
						logger.debug("Class\t", name, "\timplements\t",
								interfaceName, "\tYES");
						try {
                            
                            if (!name.endsWith("BatchDecisionResultsList"))
							rules.add(new ToxTreePackageEntry(name,files[i].getAbsolutePath(),""));
						} catch (Exception x) {
							logger.error(x);							
						}
					} else
						logger.debug("Class\t", name, "\timplements\t",
								interfaceName, "\tNO");

				}

			} catch (IOException x) {
				logger.error(x);
			}
		return rules;

	}

	/**
	 * Returns a list with available rule names.
	 * Uses {@link #getAvailableTypes(ClassLoader, String)}
	 * @return an {@link ArrayList} of class names that implement
	 *         {@link IDecisionRule}
	 */
	public static ToxTreePackageEntries getAvailableRuleTypes(ClassLoader classLoader) {
		return getAvailableTypes(classLoader,"toxTree.core.IDecisionRule");
	}

	/**
	 * Returns a list with available decision trees.
	 * Uses {@link #getAvailableTypes(ClassLoader, String)}
	 * @return an {@link ArrayList} of class names that implement
	 *         {@link IDecisionMethod}
	 */
	public static ToxTreePackageEntries getAvailableTreeTypes(ClassLoader classLoader) {
		return getAvailableTypes(classLoader,"toxTree.core.IDecisionMethod");
	}
	/**
	 * This method is the core of {@link toxTree.apps.ToxTreeApp} extension mechanism.
	 * Looks for classes that implement interface given by interfacename parameter. 
	 * Returns a list with class names.
	 * Jar files within the following directories are analyzed: <ul>
	 * <li>The directory from where the application was started (current directory "."). This is 
	 * usually the directory where toxTree.jar resides.
	 * <li>The directory "dist" below the current directory.
	 * <li>The directory "ext" below the current directory.
	 * </ul>
	 * @param classLoader
	 * @param interfacename  the 
	 * @return a list with available class names , implementing the interface{@link ArrayList}
	 */
	public static ToxTreePackageEntries getAvailableTypes(ClassLoader classLoader,String interfacename) {
		setLoader(classLoader);
		ToxTreePackageEntries listAll = null;
		for (int i=0;i < defaultLocation.length;i++) {
			try {
				ToxTreePackageEntries list = implementInterface(classLoader, 
						new File(defaultLocation[i]),
						interfacename);
				if (list.size() != 0) {
					if (listAll == null) listAll = new ToxTreePackageEntries();
					listAll.addAll(list);
				}
			} catch (toxTree.exceptions.IntrospectionException x) {
				x.printStackTrace();
			}
		}		
		return listAll;

	}	
	/**
	 * Returns a list with available categories
	 * 
	 * @return an {@link ArrayList} of class names that implement
	 *         {@link IDecisionCategory}
	 */
	public static ToxTreePackageEntries getAvailableCategoryTypes(ClassLoader classLoader) {
		return getAvailableTypes(classLoader,"toxTree.core.IDecisionCategory");
	}

	/**
	 * Creates object given the class name
	 * 
	 * @param className
	 * @return Object 
	 */
	public static Object createObject(String className) {
		Object object = null;
		try {
			Class classDefinition = Class.forName(className);
			object = classDefinition.newInstance();
		} catch (InstantiationException e) {
			logger.error(e);
		} catch (IllegalAccessException e) {
			logger.error(e);
		} catch (ClassNotFoundException e) {
			logger.error(e);
		}
		return object;
	}

	public static Object loadCreateObject(String className) {
		if (loader == null)
			return createObject(className);
		Object object = null;
		try {
			Class classDefinition = loader.loadClass(className);
			object = classDefinition.newInstance();
		} catch (InstantiationException e) {
			e.printStackTrace();
			logger.error(e);
		} catch (IllegalAccessException e) {
			e.printStackTrace();
			logger.error(e);
		} catch (ClassNotFoundException e) {
			e.printStackTrace();
			logger.error(e);
		}
		return object;
	}

	public static ClassLoader getLoader() {
		return loader;
	}

	public static void setLoader(ClassLoader classLoader) {
		if (loader == null) {
			URL[] defaultURL = new URL[defaultLocation.length];
			for (int i = 0; i < defaultURL.length; i++) {
				try {

					 //URL u = new URL("jar", "", plugins[i] + "!/");
					 defaultURL[i] = new URL("jar", "",defaultLocation[i] + "!/");
					//defaultURL[i] = new URL(defaultLocation[i]);

				} catch (MalformedURLException x) {
					x.printStackTrace();
				}
			}
			loader = new ToxTreeClassLoader(defaultURL, classLoader);
		}
	}
	/**
	 * Loads a decision tree from an InputStream
	 * Uses Java serialization mechanism
	 * @param stream
	 * @param newTitle
	 * @return a decision tree {@link IDecisionMethod}
	 */
    public static IDecisionMethod loadRules(InputStream stream, String newTitle) {
		try {
			
		    ToxTreeObjectInputStream in = new ToxTreeObjectInputStream(stream);
		    
		    IDecisionMethod tree  = (IDecisionMethod) in.readObject();
		    tree.setModified(false);
		    if (tree != null)  {
		    	tree.setExplanation(tree.getExplanation() + "\n" + newTitle);
			    		    }
    		JOptionPane.showMessageDialog(null,
    				tree.toString(),
				    "Decision tree loaded successfully.",						
				    JOptionPane.INFORMATION_MESSAGE);
    		return tree;
		} catch (Exception e) {
			e.printStackTrace();
			
			JOptionPane.showMessageDialog(null,
					e.getMessage(),
				    "Error when loading the decision tree from",						
				    JOptionPane.INFORMATION_MESSAGE);
			return null;
		}
   }    


    public static IDecisionMethodsList loadForest(InputStream stream) {
		try {
			
		    ToxTreeObjectInputStream in = new ToxTreeObjectInputStream(stream);
		    
		    IDecisionMethodsList methods  = (IDecisionMethodsList) in.readObject();
		    
    		JOptionPane.showMessageDialog(null,
    				methods.size() + " decision trees",
				    "Decision forest loaded successfully.",						
				    JOptionPane.INFORMATION_MESSAGE);
    		return methods;
		} catch (Exception e) {
			e.printStackTrace();
			
			JOptionPane.showMessageDialog(null,
					e.getMessage(),
				    "Error when loading the decision forest from",						
				    JOptionPane.INFORMATION_MESSAGE);
			return null;
		}
   }     
    
    public static IDecisionMethod loadRulesXML(InputStream stream, String newTitle) {
		try {
			Thread.currentThread().setContextClassLoader(loader);

			XMLDecoder decoder = new XMLDecoder(stream);
			IDecisionMethod tree = (IDecisionMethod)decoder.readObject();
			decoder.close();
		    
		    if (tree != null)  {
		    	tree.setExplanation(tree.getExplanation() + "\n" + newTitle);
		    	tree.setModified(false);
			    		    }
    		JOptionPane.showMessageDialog(null,
    				tree.toString(),
				    "Decision tree loaded successfully.",						
				    JOptionPane.INFORMATION_MESSAGE);
    		return tree;
		} catch (Exception e) {
			e.printStackTrace();
			
			JOptionPane.showMessageDialog(null,
					e.getMessage(),
				    "Error when loading the decision tree from",						
				    JOptionPane.INFORMATION_MESSAGE);
			return null;
		}
   }  
    public static IDecisionMethodsList loadForestXML(InputStream stream) {
		try {
			
			Thread.currentThread().setContextClassLoader(loader);

			XMLDecoder decoder = new XMLDecoder(stream);
			IDecisionMethodsList methods = (IDecisionMethodsList)decoder.readObject();
			decoder.close();
		    
    		JOptionPane.showMessageDialog(null,
    				methods.size() + " decision trees",
				    "Decision forest loaded successfully.",						
				    JOptionPane.INFORMATION_MESSAGE);
    		return methods;
		} catch (Exception e) {
			e.printStackTrace();
			
			JOptionPane.showMessageDialog(null,
					e.getMessage(),
				    "Error when loading the decision forest from",						
				    JOptionPane.INFORMATION_MESSAGE);
			return null;
		}
   }          
}
